// ---------------------------------------------------------------------------
//
//           The general purpose cross platform C/C++ framework
//
//                       S x A c c e l e r a t e
//
//           Home:       https://www.sxlib.de
//           License:    Apache 2
//           Authors:    see src/AUTHORS
//
// ---------------------------------------------------------------------------

#include <SxSymbol.h>

namespace SxParserKit {

SxSymbol::SxSymbol ()
   : name (),
     type (Type::Undefined),
     dataG(),
     value (NULL),
     nodeId (-1)
{
   // empty
}

SxSymbol::SxSymbol (const SxPtr<SxGraph<SxGProps> > &gPtr_,
                    ssize_t nodeId_)
{
   SX_CHECK (gPtr_.getPtr ());

   dataG   = gPtr_;
   value   = NULL;
   nodeId  = nodeId_;

   auto it = dataG->begin (nodeId);

   SX_CHECK (it.isValid ());

   name  = it->getProperty (".key").toString ();
   value = &(it->getProperty (".val"));
   type  = (Type)it->getProperty (".type").getInt ();
}

SxSymbol::~SxSymbol ()
{
   // empty
}

SxString SxSymbol::toString () const
{
   SX_TRACE ();

   if (type != Type::String)
      SX_THROW (  "SxSymbol", "IOError",
                  value->getTag () + ": Element of type '"
                + SxParserAst::getTypeStr (type)
                + "' cannot be converted to 'string'");

   return value->toString ();
}

int64_t SxSymbol::toInt () const
{
   SX_TRACE ();

   bool isAllowed = (type == Type::Int || type == Type::Double);

   if (!isAllowed)
      SX_THROW (  "SxSymbol", "IOError",
                  value->getTag () + ": Element '"
                + name + "' of type '"
                + SxParserAst::getTypeStr (type)
                + "' cannot be converted to 'int'");

   return value->toInt ();
}

double SxSymbol::toDouble () const
{
   SX_TRACE ();

   bool isAllowed = (type == Type::Int || type == Type::Double);

   if (!isAllowed)
      SX_THROW (  "SxSymbol", "IOError",
                  value->getTag () + ": Element '"
                + name + "' of type '"
                + SxParserAst::getTypeStr (type)
                + "' cannot be converted to 'double'");

   return value->toDouble ();
}

bool SxSymbol::toBool () const
{
   SX_TRACE ();

   if (type != Type::Bool)
      SX_THROW (  "SxSymbol", "IOError",
                  value->getTag () + ": Element '"
                + name + "' of type '"
                + SxParserAst::getTypeStr (type)
                + "' cannot be converted to 'bool'");

   return value->toBool ();
}

SxList<int64_t> SxSymbol::toIntList () const
{
   SX_TRACE ();

   if (type != Type::List)
      SX_THROW (  "SxSymbol", "IOError",
                  value->getTag ()
                + ": expected type 'list', got '"
                + SxParserAst::getTypeStr (type) + "'");

   SxList<int64_t> res;
   auto it = dataG->begin (nodeId);
   for (ssize_t i = 0; i < it.getSizeOut (); ++i)  {
      auto chIt = it.out (i);
      int nType = (int)chIt->getProperty (".type").getInt ();
      if (nType == Type::Double || nType == Type::Int)  {
         res.append (chIt->getProperty (".val").toInt ());
      }
   }
   return res;
}

SxList<SxString> SxSymbol::toStringList () const
{
   SX_TRACE ();

   if (type != Type::List)
      SX_THROW (  "SxSymbol", "IOError",
                  value->getTag ()
                + ": expected type 'list', got '"
                + SxParserAst::getTypeStr (type) + "'");

   SxList<SxString> res;
   auto it = dataG->begin (nodeId);
   for (ssize_t i = 0; i < it.getSizeOut (); ++i)  {
      auto chIt = it.out (i);
      int nType = (int)chIt->getProperty (".type").getInt ();
      if (nType == Type::String)  {
         res.append (chIt->getProperty (".val").toString ());
      }
   }
   return res;
}

SxList<double> SxSymbol::toDoubleList () const
{
   SX_TRACE ();

   if (type != Type::List)
      SX_THROW (  "SxSymbol", "IOError",
                  value->getTag ()
                + ": expected type 'list', got '"
                + SxParserAst::getTypeStr (type) + "'");

   SxList<double> res;
   auto it = dataG->begin (nodeId);
   for (ssize_t i = 0; i < it.getSizeOut (); ++i)  {
      auto chIt = it.out (i);
      int nType = (int)chIt->getProperty (".type").getInt ();
      if (nType == Type::Double || nType == Type::Int)  {
         res.append (chIt->getProperty (".val").toDouble ());
      }
   }
   return res;
}

SxList<SxSymbol> SxSymbol::toList () const
{
   SX_TRACE ();

   if (type != Type::List && type != Type::Group)
      SX_THROW (  "SxSymbol", "IOError",
                  value->getTag ()
                + ": expected type 'list' or 'group', got '"
                + SxParserAst::getTypeStr (type) + "'");

   SxList<SxSymbol> res;
   // --- collect all child elements
   auto it = dataG->begin (nodeId);
   for (ssize_t i = 0; i < it.getSizeOut (); ++i)  {
      auto chIt = it.out (i);
      res.append (SxSymbol(dataG, chIt.getIdx ()));
   }

   return res;
}

bool SxSymbol::hasElem (const SxString &key) const
{
   SX_TRACE ();

   if (type != Type::Group)
      SX_THROW (  "SxSymbol", "IOError",
                  value->getTag ()
                + ": expected type 'group', got '"
                + SxParserAst::getTypeStr (type) + "'");

   auto it = dataG->begin (nodeId);
   // --- search within all child elements
   for (ssize_t i = 0; i < it.getSizeOut (); ++i)  {
      auto chIt = it.out (i);
      if (chIt->getProperty (".key").toString () == key)  {
         return true;
      }
   }

   return false;
}

SxSymbol SxSymbol::getElem (const SxString &key, bool isRequired) const
{
   SX_TRACE ();

   if (type != Type::Group)
      SX_THROW (  "SxSymbol", "IOError",
                  value->getTag ()
                + ": expected type 'group', got '"
                + SxParserAst::getTypeStr (type) + "'");

   if (isRequired && !hasElem (key))
      SX_THROW (  "SxSymbol", "IOError",
                  value->getTag ()
                + ": Element with key '"+ key
                + "' not found in group '" + name + "'");

   auto it = dataG->begin (nodeId);
   // --- search within all child elements
   for (ssize_t i = 0; i < it.getSizeOut (); ++i)  {
      auto chIt = it.out (i);
      if (chIt->getProperty (".key").toString () == key)  {
         SxSymbol s(dataG, chIt.getIdx ());
         return s;
      }
   }
   return SxSymbol ();
}

SxList<SxSymbol> SxSymbol::getTypedList (Type type_) const
{
   SX_TRACE ();

   if (type != Type::List && type != Type::Group)
      SX_THROW (  "SxSymbol", "IOError",
                  value->getTag ()
                + ": expected type 'list' or 'group', got '"
                + SxParserAst::getTypeStr (type) + "'");

   SxList<SxSymbol> res;

   auto it = dataG->begin (nodeId);
   for (ssize_t i = 0; i < it.getSizeOut (); ++i)  {
      auto chIt = it.out (i);
      SxSymbol tr (dataG, chIt.getIdx ());
      if (tr.getType () == type_)  res.append (tr);
   }
   return res;
}

SxList<SxSymbol>
SxSymbol::getTypedList (const SxString &key_, Type type_) const
{
   SX_TRACE ();

   if (type != Type::Group)
      SX_THROW (  "SxSymbol", "IOError",
                  value->getTag ()
                + ": expected type 'group', got '"
                + SxParserAst::getTypeStr (type) + "'");

   SxList<SxSymbol> res;

   auto it = dataG->begin (nodeId);
   for (ssize_t i = 0; i < it.getSizeOut (); ++i)  {
      auto chIt = it.out (i);
      SxSymbol tr (dataG, chIt.getIdx ());
      if (tr.getType () == type_ && tr.getName () == key_)  {
         res.append (tr);
      }
   }

   return res;
}

SxList<SxSymbol> SxSymbol::getGroupList () const
{
   SX_TRACE ();
   return getTypedList (Type::Group);
}

SxList<SxSymbol> SxSymbol::getGroupList (const SxString &key_) const
{
   SX_TRACE ();
   return getTypedList (key_, Type::Group);
}

SxList<SxSymbol> SxSymbol::getArrayList () const
{
   SX_TRACE ();
   return getTypedList (Type::List);
}

SxList<SxSymbol> SxSymbol::getArrayList (const SxString &key_) const
{
   SX_TRACE ();
   return getTypedList (key_, Type::List);
}

SxString SxSymbol::getName () const
{
   SX_TRACE ();
   return name;
}

SxSymbol::Type SxSymbol::getType () const
{
   SX_TRACE ();
   return type;
}

SxString SxSymbol::getTag () const
{
   SX_TRACE ();
   return value->getTag ();
}

bool SxSymbol::isValid () const
{
   SX_TRACE ();
   if (type != Type::Undefined)  return true;
   else                          return false;
}

void SxSymbol::print (ostream &os, const SxString &newLine,
                      int lvl, int tabs) const
{
   SX_TRACE ();
   SX_CHECK (isValid ());
   SX_UNUSED (newLine);
   if (type == Type::Group)  {
      printObject (os, newLine, lvl, tabs);
   }  else if (type == Type::List)  {
      printArray (os, newLine, lvl, tabs);
   }  else if (type == Type::Int)  {
      printInt (os, lvl, tabs);
   }  else if (type == Type::Double)  {
      printDouble (os, lvl, tabs);
   }  else if (type == Type::String)  {
      printString (os, lvl, tabs);
   }  else if (type == Type::Bool)  {
      printBool (os, lvl, tabs);
   }  else  {
      SX_EXIT;
   }
}

void SxSymbol::printObject (ostream &os, const SxString &newLine,
                            int lvl, int tabs) const
{
   SX_TRACE ();
   SX_CHECK (type == Type::Group, type);
   SxList<SxSymbol> elems = toList ();
   ssize_t c = 0;
   for (int i = 0; i < lvl*tabs; ++i)  os << " ";

   if (name != "")  os << "\"" << name << "\":";

   os << "{";
   for (auto it = elems.begin (); it != elems.end (); ++it)
   {
      if (c != 0)  os << ",";
      os << newLine;
      it->print (os, newLine, lvl+1, tabs);
      c++;
   }
   os << newLine;
   for (int i = 0; i < lvl*tabs; ++i)  os << " ";
   os << "}";
}

void SxSymbol::printArray (ostream &os, const SxString &newLine,
                           int lvl, int tabs) const
{
   SX_TRACE ();
   SX_CHECK (type == Type::List, type);
   SxList<SxSymbol> elems = toList ();
   ssize_t c = 0;
   for (int i = 0; i < lvl*tabs; ++i)  os << " ";

   if (name != "")  os << "\"" << name << "\":";

   os << "[";
   for (auto it = elems.begin (); it != elems.end (); ++it)
   {
      if (c != 0)  os << ",";
      os << newLine;
      it->print (os, newLine, lvl+1, tabs);
      c++;
   }
   os << newLine;
   for (int i = 0; i < lvl*tabs; ++i)  os << " ";
   os << "]";

}

void SxSymbol::printInt (ostream &os, int lvl, int tabs) const
{
   SX_TRACE ();
   SX_CHECK (type == Type::Int, type);

   for (int i = 0; i < lvl*tabs; ++i)  os << " ";

   if (name != "")  os << "\"" << name << "\":";

   os << toInt ();
}

void SxSymbol::printDouble (ostream &os, int lvl, int tabs) const
{
   SX_TRACE ();
   SX_CHECK (type == Type::Double, type);

   for (int i = 0; i < lvl*tabs; ++i)  os << " ";

   if (name != "")  os << "\"" << name << "\":";

   os << toDouble ();
}

void SxSymbol::printString (ostream &os, int lvl, int tabs) const
{
   SX_TRACE ();
   SX_CHECK (type == Type::String, type);

   for (int i = 0; i < lvl*tabs; ++i)  os << " ";

   if (name != "")  os << "\"" << name << "\":";

   os << "\"" << toString () << "\"";
}

void SxSymbol::printBool (ostream &os, int lvl, int tabs) const
{
   SX_TRACE ();
   SX_CHECK (type == Type::Bool, type);

   for (int i = 0; i < lvl*tabs; ++i)  os << " ";

   if (name != "")  os << "\"" << name << "\":";

   if (toBool ())  os << "true";
   else            os << "false";
}

} /* SxParserKit */
