// ---------------------------------------------------------------------------
//
//           The general purpose cross platform C/C++ framework
//
//                       S x A c c e l e r a t e
//
//           Home:       https://www.sxlib.de
//           License:    Apache 2
//           Authors:    see src/AUTHORS
//
// ---------------------------------------------------------------------------

#ifndef _SX_SYMBOL_H_
#define _SX_SYMBOL_H_

#include <SxJSON.h>
#include <SxException.h>
#include <SxParserAst.h>
#include <SxGraph.h>
#include <SxGProps.h>
#include <SxGQuery.h>

/** \brief Reader for tree structured data

    \b SxSymbol = SPHInX Tree format Symbol

     SxSymbol class provides access
     functions to retrieve data out of
     file stored in tree structure that
     is represented by SxGraph.
 */

SX_EXCEPTION_CAT ("SxSymbol");
SX_EXCEPTION_TAG ("IOError");

SX_EXCEPTION_TYPE ("SxSymbol", "IOError", "message", SxString);

namespace SxParserKit {
class SX_EXPORT_JSON SxSymbol
{
   public:

      typedef typename SxGQExprBase::SelSet     SelSet;
      typedef typename SxParserAst::ElemType    Type;

      SxSymbol ();
      SxSymbol (const SxPtr<SxGraph<SxGProps> > &gPtr_, ssize_t nodeId_);

     ~SxSymbol ();

      // --- All to.. functions convert the current SxSymbol
      //     object to a specific type.
      SxString toString () const;
      int64_t  toInt () const;
      double   toDouble () const;
      bool     toBool () const;

      // --- functions to convert the SxSymbol
      //     object of type List to SxList of
      //     specific type.
      SxList<int64_t>   toIntList ()    const;
      SxList<SxString>  toStringList () const;
      SxList<double>    toDoubleList () const;
      SxList<SxSymbol>  toList ()       const;

      // --- check existence an element
      bool hasElem (const SxString &key) const;

      // --- retrieve elem with given key
      //     inside the current SxSymbol object.
      SxSymbol getElem (const SxString &key, bool isRequired=true) const;


      // --- retrieve all elements of Group type
      SxList<SxSymbol> getGroupList () const;
      // --- retrieve all elements of Group type and given 'key_'
      SxList<SxSymbol> getGroupList (const SxString &key_) const;
      // --- retrieve all elements of List type
      SxList<SxSymbol> getArrayList () const;
      // --- retrieve all elements of List type and given 'key_'
      SxList<SxSymbol> getArrayList (const SxString &key_) const;

      SxString getName () const;
      Type     getType () const;
      SxString getTag  () const;

      bool isValid () const;
      void print (ostream &os=cout, const SxString &newLine="\n",
                  int lvl = 0, int tabs = 3) const;

   protected:

      // --- retrieve all elements of given type
      SxList<SxSymbol> getTypedList (Type type_) const;
      SxList<SxSymbol> getTypedList (const SxString &key_, Type type_) const;

      // --- print in JSON format
      void printObject (ostream &os, const SxString &newLine,
                        int lvl, int tabs = 3) const;
      void printArray (ostream &os, const SxString &newLine,
                       int lvl, int tabs = 3) const;
      void printInt (ostream &os, int lvl, int tabs = 3) const;
      void printDouble (ostream &os, int lvl, int tabs = 3) const;
      void printString (ostream &os, int lvl, int tabs = 3) const;
      void printBool (ostream &os, int lvl, int tabs = 3) const;

      SxString name; // key
      Type type; // type of value
      SxPtr<SxGraph<SxGProps> > dataG;
      SxVariant *value;
      ssize_t nodeId; // graph node Id
};

} /* SxParserKit */

#endif /* _SX_SYMBOL_H_ */
