#include <SxSchema.h>

SxSchema::SxSchema () { }

SxSchema::~SxSchema () { }

SxSchema::SxSchema (const SxPtr<SxGraph<SxGProps> > &schemaG_)
{
   SX_CHECK (schemaG_.getPtr ());

   schemaG  = schemaG_;
   simplifySchemaAst ();
}

bool SxSchema::getBool (const SxGraph<SxGProps>::Iterator &sIt,
                        const SxString &key,
                        bool default_)
{
   auto it = get (sIt, ".key", key);
   if (it.isValid ())
      return it->getProperty (".val").toBool ();
   else
      return default_;
}

int64_t SxSchema::getInt (const SxGraph<SxGProps>::Iterator &sIt,
                          const SxString &key,
                          const int64_t &default_)
{
   auto it = get (sIt, ".key", key);
   if (it.isValid ())
      return it->getProperty (".val").toInt ();
   else
      return default_;
}

double SxSchema::getDouble (const SxGraph<SxGProps>::Iterator &sIt,
                            const SxString &key,
                            const double &default_)
{
   auto it = get (sIt, ".key", key);
   if (it.isValid ())
      return it->getProperty (".val").toDouble ();
   else
      return default_;
}

SxString SxSchema::getString (const SxGraph<SxGProps>::Iterator &sIt,
                              const SxString &key,
                              const SxString &default_)
{
   auto it = get (sIt, ".key", key);
   if (it.isValid ())
      return it->getProperty (".val").toString ();
   else
      return default_;
}

void SxSchema::validationError (const SxString &msg,
                                const SxString &sTag,
                                const SxString &dTag)
{
   SX_THROW ("SchemaValidation", "ValidationError",
             dTag+": "+msg+", declared at "+sTag);
}

int SxSchema::getRank (const SxGraph<SxGProps>::Iterator &dIt)  {
   SX_CHECK (dIt.isValid ());

   int maxRank = 0, rank = 0;
   for (ssize_t i = 0; i < dIt.getSizeOut (); ++i) {
      auto eIt = dIt.out (i);
      if (eIt->getProperty (".type").getInt () == Type::List)  {
         rank = getRank (eIt);
         if (rank > maxRank)  maxRank = rank;
      }
   }
   return 1+maxRank;
}

bool SxSchema::validateArray (const SxGraph<SxGProps>::Iterator &sIt,
                              const SxGraph<SxGProps>::Iterator &dIt)
{
   SX_CHECK (sIt.isValid ());

   // --- find rank
   int64_t rank = getRank (dIt);

   int64_t minRank = getInt (sIt, "minRank", min<int64_t>());
   int64_t maxRank = getInt (sIt, "maxRank", max<int64_t>());

   if (minRank != min<int64_t>() && rank < minRank)  {
      validationError (  SxString ("rank must be greater or equal to ")
                       + minRank + ", got " + rank,
                       sIt->getProperty (".val").getTag (),
                       dIt->getProperty (".val").getTag ());
   }

   if (maxRank != max<int64_t>() && rank > maxRank)  {
      validationError (  SxString ("rank must be less or equal to ")
                       + maxRank + ", got " + rank,
                       sIt->getProperty (".val").getTag (),
                       dIt->getProperty (".val").getTag ());
   }

   return true;
}

bool SxSchema::validateVector (const SxGraph<SxGProps>::Iterator &sIt,
                               const SxGraph<SxGProps>::Iterator &dIt)
{
   SX_CHECK (sIt.isValid ());

   ssize_t nElems = dIt.getSizeOut ();

   int64_t dim = getInt (sIt, "dim", max<int64_t>());

   if (dim != max<int64_t>() && dim != static_cast<int64_t>(nElems))  {
      validationError (  SxString("dimensions mismatch for vector, ")
                       + "must be equal to " + dim,
                       sIt->getProperty (".val").getTag (),
                       dIt->getProperty (".val").getTag ());
   }

   int64_t minDim = getInt (sIt, "minDim", min<int64_t>());

   if (minDim != min<int64_t>() && static_cast<int64_t>(nElems) < minDim)  {
      validationError (  SxString("dimensions mismatch for vector, ")
                       + "must be greater or equal to "
                       + minDim + ", got " + nElems,
                       sIt->getProperty (".val").getTag (),
                       dIt->getProperty (".val").getTag ());
   }

   int64_t maxDim = getInt (sIt, "maxDim", max<int64_t>());

   if (maxDim != max<int64_t>() && static_cast<int64_t>(nElems) > maxDim)  {
      validationError (  SxString("dimensions mismatch for vector, ")
                       + "must be less or equal to "
                       + maxDim + ", got " + nElems,
                       sIt->getProperty (".val").getTag (),
                       dIt->getProperty (".val").getTag ());
   }

   return true;
}

bool SxSchema::validateMatrix (const SxGraph<SxGProps>::Iterator &sIt,
                               const SxGraph<SxGProps>::Iterator &dIt)
{
   SX_CHECK (sIt.isValid ());


   int64_t dimX = max<int64_t>(), dimY = max<int64_t>();

   auto dimsIt = get (sIt, ".key", "dims");
   SX_CHECK (dimsIt.isValid ());
   SX_CHECK (dimsIt->getProperty (".type").getInt () == Type::List);

   if (   dimsIt.isValid ()
       && dimsIt->getProperty (".type").getInt () == Type::List)  {

      if (dimsIt.getSizeOut () > 0)
         dimX = dimsIt.out (0)->getProperty (".val").toInt ();

      if (dimsIt.getSizeOut () > 1)
         dimY = dimsIt.out (1)->getProperty (".val").toInt ();

      SX_CHECK (dimX != 0 && dimY != 0);

      ssize_t nElems = dIt.getSizeOut ();

      if (dimX != max<int64_t>() && dimX != static_cast<int64_t>(nElems))  {
         validationError (  SxString("number of rows ")
                          + "of matrix "
                          + "must be equal to " + dimX
                          + ", got " + nElems,
                          sIt->getProperty (".val").getTag (),
                          dIt->getProperty (".val").getTag ());
      }

      nElems = 0;

      for (ssize_t i = 0; i < dIt.getSizeOut (); ++i)  {
         auto eIt = dIt.out (i);
         SX_CHECK (eIt->getProperty (".type").toInt () == Type::List);
         nElems = eIt.getSizeOut ();
         if (dimY != static_cast<int64_t>(nElems))  {
            validationError (  SxString("number of columns ")
                             + " of matrix "
                             + "must be equal to " + dimY
                             + ", got " + nElems,
                             sIt->getProperty (".val").getTag (),
                             eIt->getProperty (".val").getTag ());
         }
      }
   }
   return true;
}

bool SxSchema::validateObj (const SxGraph<SxGProps>::Iterator &sIt,
                            SxGraph<SxGProps>::Iterator *dIt)
{
   SX_CHECK (dIt);
   SX_CHECK (sIt.isValid ());

   auto dataIt = *dIt;

   // --- any unexpected items
   if (!allowUndeclared)  {
      for (ssize_t m = 0; m < dataIt.getSizeOut (); ++m)  {
         bool found = false;
         SxString key = dataIt.out (m)->getProperty (".key").getString ();
         for (ssize_t n = 0; n < sIt.getSizeOut (); ++n)  {
            SxString k = sIt.out (n)->getProperty (".key").getString ();
            if (key == k)  {
               found = true;
               break;
            }
         }
         if (!found)  {
            validationError (  "undeclared key '"
                             + key + "' in group '"
                             + sIt->getProperty (".key").getString ()
                             + "'",
                             sIt->getProperty (".val").getTag (),
                             dataIt.out (m)->getProperty (".val").getTag ());
         }
      }
   }

   SxMap<ssize_t, SxList<ssize_t> > foundGroups;
   SxMap<SxString, ssize_t> nameToIdx;
   SxArray<uint8_t> isFound;
   isFound.resize (sIt.getSizeOut ());
   if (isFound.getSize () > 0)  isFound.set (0);
   for (ssize_t i = 0; i < sIt.getSizeOut (); ++i)  {
      auto chIt = sIt.out (i);

      if (chIt->getProperty (".type").getInt () != Type::Group)
         continue;

      Type type = (Type)getInt (chIt, "type", (int)Type::Undefined);

      const SxString &key = chIt->getProperty (".key").getString ();

      nameToIdx (key) = i;

      if (type == Type::Group)  {
         foundGroups(chIt.getIdx ()); // insert empty matched list

         for (ssize_t k = 0; k < dataIt.getSizeOut (); ++k)  {
            auto dchIt = dataIt.out (k);
            if (dchIt->getProperty (".key").getString () == key)  {
               if (dchIt->getProperty (".type").getInt () != Type::Group)  {
                  Type t = (Type)dchIt->getProperty (".type").getInt ();
                  SxString typeStr = SxParserAst::getTypeStr (t);
                  validationError (  "type mismatch for key '" + key
                                   + "', expected 'group', got '"
                                   + typeStr + "'",
                                   dchIt->getProperty (".val").getTag (),
                                   chIt->getProperty (".val").getTag ());
               }

               foundGroups(chIt.getIdx ()).append (dchIt.getIdx ());
               isFound(i) = 1;
            }
         }
      }  else if (type == Type::List)  {
         ssize_t prevId = -1;
         for (ssize_t k = 0; k < dataIt.getSizeOut (); ++k)  {
            auto dchIt = dataIt.out (k);
            if (dchIt->getProperty (".key").getString () == key)  {
               if (prevId != -1)  {
                  validationError (  "duplicate key '"
                                   + key
                                   + "' previously defined at "
                                   + dataIt.out (prevId)->getProperty (".val").getTag (),
                                   chIt->getProperty (".val").getTag (),
                                   dchIt->getProperty (".val").getTag ());
               }

               if (dchIt->getProperty (".type").getInt () != Type::List)  {
                  Type t = (Type)dchIt->getProperty (".type").getInt ();
                  SxString typeStr = SxParserAst::getTypeStr (t);
                  validationError (  "type mismatch for key '" + key
                                   + "', expected 'list', got '"
                                   + typeStr + "'",
                                   dchIt->getProperty (".value").getTag (),
                                   chIt->getProperty (".val").getTag ());
               }
               prevId = k;
               isFound(i) = 1;
               validateArray (chIt, dchIt);
            }
         }

      }  else if (type == Type::Vector)  {
         ssize_t prevId = -1;
         for (ssize_t k = 0; k < dataIt.getSizeOut (); ++k)  {
            auto dchIt = dataIt.out (k);
            if (dchIt->getProperty (".key").getString () == key)  {
               if (prevId != -1)  {
                  validationError (  "duplicate key '"
                                   + key
                                   + "' previously defined at "
                                   + dataIt.out (prevId)->getProperty (".val").getTag (),
                                   chIt->getProperty (".val").getTag (),
                                   dchIt->getProperty (".val").getTag ());
               }

               if (dchIt->getProperty (".type").getInt () != Type::List)  {
                  Type t = (Type)dchIt->getProperty (".type").getInt ();
                  SxString typeStr = SxParserAst::getTypeStr (t);
                  validationError (  "type mismatch for key '" + key
                                   + "', expected 'list(vector)', got '"
                                   + typeStr + "'",
                                   dchIt->getProperty (".val").getTag (),
                                   chIt->getProperty (".val").getTag ());
               }
               prevId = k;
               isFound(i) = 1;
               validateVector (chIt, dchIt);
            }
         }

      }  else if (type == Type::Matrix)  {
         ssize_t prevId = -1;
         for (ssize_t k = 0; k < dataIt.getSizeOut (); ++k)  {
            auto dchIt = dataIt.out (k);
            if (dchIt->getProperty (".key").getString () == key)  {
               if (prevId != -1)  {
                  validationError (  "duplicate key '"
                                   + key
                                   + "' previously defined at "
                                   + dataIt.out (prevId)->getProperty (".val").getTag (),
                                   chIt->getProperty (".val").getTag (),
                                   dchIt->getProperty (".val").getTag ());
               }

               if (dchIt->getProperty (".type").getInt () != Type::List)  {
                  Type t = (Type)dchIt->getProperty (".type").getInt ();
                  SxString typeStr = SxParserAst::getTypeStr (t);
                  validationError (  "type mismatch for key '" + key
                                   + "', expected 'list(matrix)', got '"
                                   + typeStr + "'",
                                   dchIt->getProperty (".val").getTag (),
                                   chIt->getProperty (".val").getTag ());
               }
               prevId = k;
               isFound(i) = 1;
               validateMatrix (chIt, dchIt);
            }
         }

      }  else if (type == Type::Enum)  {
         ssize_t prevId = -1;
         for (ssize_t k = 0; k < dataIt.getSizeOut (); ++k)  {
            auto dchIt = dataIt.out (k);
            if (dchIt->getProperty (".key").getString () == key)  {
               isFound(i) = 1;
               if (prevId != -1)  {
                  validationError (  "duplicate key '"
                                   + key
                                   + "' previously defined at "
                                   + dataIt.out (prevId)->getProperty (".val").getTag (),
                                   chIt->getProperty (".val").getTag (),
                                   dchIt->getProperty (".val").getTag ());
               }

               if (   dchIt->getProperty (".val").getType ()
                   != SxVariantType::String)
               {
                  validationError (  "type mismatch for key '" + key
                                   + "', expected 'string(enum)', got '"
                                   + dchIt->getProperty (".val").getTypeName ()
                                   + "'",
                                   chIt->getProperty (".val").getTag (),
                                   dchIt->getProperty (".val").getTag ());
               }

               SxString val = dchIt->getProperty (".val").getString ();

               SxList<SxString> options = getString (chIt, "val", "").tokenize(",");
               bool found = false;
               for (auto oIt = options.begin (); oIt.isValid (); ++oIt)  {
                  if (*oIt == val)  {
                     found = true;
                     break;
                  }
               }

               if (found == false)  {
                  validationError (  SxString ("value for key '"+ key
                                   + "' not part of Enum values"),
                                   chIt->getProperty (".val").getTag (),
                                   dchIt->getProperty (".val").getTag ());
               }
               prevId = k;
            }
         }

      }  else if (type == Type::Int)  {
         ssize_t prevId = -1;
         for (ssize_t k = 0; k < dataIt.getSizeOut (); ++k)  {
            auto dchIt = dataIt.out (k);
            if (dchIt->getProperty (".key").getString () == key)  {
               isFound(i) = 1;
               if (prevId != -1)  {
                  validationError (  "duplicate key '"
                                   + key
                                   + "' previously defined at "
                                   + dataIt.out (prevId)->getProperty (".val").getTag (),
                                   chIt->getProperty (".val").getTag (),
                                   dchIt->getProperty (".val").getTag ());
               }

               const SxVariant &val = dchIt->getProperty (".val");

               if (   val.getType () != SxVariantType::Int
                   && val.getType () != SxVariantType::Double)  {
                  validationError (  "type mismatch for key '" + key
                                   + "', expected 'int', got '"
                                   + val.getTypeName () + "'",
                                   chIt->getProperty (".val").getTag (),
                                   val.getTag ());
               }


               if (!chIt->getProperty (".val").matchLimits (val))  {
                  int64_t minVal = 0;
                  int64_t maxVal = 0;
                  chIt->getProperty (".val").getLimits (&minVal, &maxVal);
                  validationError (  SxString("int value ")
                                   + val.toInt () + " is out of range ["
                                   + minVal + ", "
                                   + maxVal + "]",
                                   chIt->getProperty (".val").getTag (),
                                   val.getTag ());
               }
               prevId = k;
            }
         }

      }  else if (type == Type::Double)  {
         ssize_t prevId = -1;
         for (ssize_t k = 0; k < dataIt.getSizeOut (); ++k)  {
            auto dchIt = dataIt.out (k);
            if (dchIt->getProperty (".key").getString () == key)  {
               isFound(i) = 1;
               if (prevId != -1)  {
                  validationError (  "duplicate key '"
                                   + key
                                   + "' previously defined at "
                                   + dataIt.out (prevId)->getProperty (".val").getTag (),
                                   chIt->getProperty (".val").getTag (),
                                   dchIt->getProperty (".val").getTag ());
               }

               const SxVariant &val = dchIt->getProperty (".val");

               if (   val.getType () != SxVariantType::Int
                   && val.getType () != SxVariantType::Double)
               {
                  validationError (  "type mismatch for key '" + key
                                   + "', expected 'real', got '"
                                   + val.getTypeName () + "'",
                                   chIt->getProperty (".val").getTag (),
                                   val.getTag ());
               }

               if (!chIt->getProperty (".val").matchLimits (val))  {
                  double minVal = 0;
                  double maxVal = 0;
                  chIt->getProperty (".val").getLimits (&minVal, &maxVal);
                  validationError (  SxString("real value ")
                                   + val.toDouble () + " is out of range ["
                                   + minVal + ", "
                                   + maxVal + "]",
                                   chIt->getProperty (".val").getTag (),
                                   val.getTag ());
               }
               prevId = k;
            }
         }

      }  else if (type == Type::Bool)  {
         ssize_t prevId = -1;
         for (ssize_t k = 0; k < dataIt.getSizeOut (); ++k)  {
            auto dchIt = dataIt.out (k);
            if (dchIt->getProperty (".key").getString () == key)  {
               isFound(i) = 1;
               if (prevId != -1)  {
                  validationError (  "duplicate key '"
                                   + key
                                   + "' previously defined at "
                                   + dataIt.out (prevId)->getProperty (".val").getTag (),
                                   chIt->getProperty (".val").getTag (),
                                   dchIt->getProperty (".val").getTag ());
               }

               SxVariant &val = dchIt->getProperty (".val");

               // --- correction for symboltable issue
               if (val.getType () == SxVariantType::Int)  {
                  int64_t v = dchIt->getProperty (".val").toInt ();
                  if (v > 0)  dchIt->setProperty (".val", true);
                  else        dchIt->setProperty (".val", false);
                  dchIt->setProperty (".type", (int)Type::Bool);
               }

               if (val.getType () != SxVariantType::Bool)  {
                  validationError (  "type mismatch for key '" + key
                                   + "', expected 'bool', got '"
                                   + val.getTypeName () + "'",
                                   chIt->getProperty (".val").getTag (),
                                   val.getTag ());
               }
               prevId = k;
            }
         }

      }  else if (type == Type::String)  {
         ssize_t prevId = -1;
         for (ssize_t k = 0; k < dataIt.getSizeOut (); ++k)  {
            auto dchIt = dataIt.out (k);
            if (dchIt->getProperty (".key").getString () == key)  {
               isFound(i) = 1;
               if (prevId != -1)  {
                  validationError (  "duplicate key '"
                                   + key
                                   + "' previously defined at "
                                   + dataIt.out (prevId)->getProperty (".val").getTag (),
                                   chIt->getProperty (".val").getTag (),
                                   dchIt->getProperty (".val").getTag ());
               }

               const SxVariant &val = dchIt->getProperty (".val");

               if (val.getType () != SxVariantType::String)
               {
                  validationError (  "type mismatch for key '" + key
                                   + "', expected 'string', got '"
                                   + val.getTypeName () + "'",
                                   chIt->getProperty (".val").getTag (),
                                   val.getTag ());
               }


               if (!chIt->getProperty (".val").matchLimits (val))  {
                  int64_t minLen = 0;
                  int64_t maxLen = 0;
                  chIt->getProperty (".val").getLimits (&minLen, &maxLen);
                  validationError (  SxString("string of size ")
                                   + val.toString ().getSize ()
                                   + " is out of range ["
                                   + minLen + ", "
                                   + maxLen + "]",
                                   chIt->getProperty (".val").getTag (),
                                   val.getTag ());
               }
               prevId = k;
            }
         }

      }  else  {
         validationError (  "type '" + SxParserAst::getTypeStr(type)
                          + "' not recognized",
                          chIt->getProperty (".val").getTag (),
                          chIt->getProperty (".val").getTag ());
      }

      bool opt = getBool (chIt, "optional", false);

      // --- if not found and is optional, set to 2
      if (isFound(i) == 0 && opt == true)  {
         isFound(i) = 2;
      }

      // ----------------------------------------------------

      if (isFound(i) == 1)  {
         // --- check for needed items
         SxString needs = getString (chIt, "needs", "");

         if (needs != "")  {
            SxList<SxString> needList = needs.tokenize (",");
            bool found = true;
            for (auto nIt = needList.begin (); nIt.isValid (); ++nIt)
            {
               if (!(get(dataIt,".key", *nIt).isValid()))  {
                  found = false;
                  break;
               }
            }

            if (!found)  {
               validationError (  "Presence of item(s) '"
                                + needs
                                + "' is required by '"
                                + chIt->getProperty (".key").getString ()
                                + "'",
                                chIt->getProperty (".val").getTag (),
                                dataIt->getProperty (".val").getTag ()
                               );
            }
         }

         // --- check for excluded items
         SxString nots = getString (chIt, "not", "");

         if (nots != "")  {
            SxList<SxString> notList = nots.tokenize (",");
            bool found = false;
            for (auto nIt = notList.begin (); nIt.isValid (); ++nIt)
            {
               if ((get(dataIt,".key", *nIt).isValid()))  {
                  found = true;
                  break;
               }
            }

            if (found)  {
               validationError (  "Presence of item(s) '"
                                + nots
                                + "' is excluded by '"
                                + chIt->getProperty (".key").getString ()
                                + "'",
                                chIt->getProperty (".val").getTag (),
                                dataIt->getProperty (".val").getTag ()
                               );
            }
         }
      }

   }

   for (ssize_t i = 0; i < sIt.getSizeOut (); ++i)  {
      auto chIt = sIt.out (i);

      if (chIt->getProperty (".type").getInt () != Type::Group)
         continue;

      const SxString &key = chIt->getProperty (".key").getString ();
      SxString xorStr = getString (chIt, "xor", "");

      if (xorStr != "")  {
         SxList<SxString> xorList = xorStr.tokenize (',');
         bool xorFound = false;
         for (auto xorIt = xorList.begin (); xorIt != xorList.end (); ++xorIt)
         {
            if (   isFound(nameToIdx (*xorIt)) == 1
                && *xorIt != key)  {
               xorFound = true;
            }
         }

         if (xorFound == false && isFound (i) == 0)  {
            validationError (  "Missing one of '"
                             + xorStr + "'",
                             chIt->getProperty (".val").getTag (),
                             dataIt->getProperty (".val").getTag ());
         }
         // --- if curr key is found and one of xor's is also found
         if (xorFound == true && isFound (i) == 1)  {
            validationError (xorStr + "' are mutually excluded",
                             chIt->getProperty (".val").getTag (),
                             dataIt->getProperty (".val").getTag ());
         }
      }  else  {
         // --- if not optional
         if (isFound (i) == 0)  {
            // --- item not found
            validationError (  "key '" + key
                             + "' not found in '"
                             + dataIt->getProperty (".key").getString ()
                             + "'", sIt->getProperty (".val").getTag (),
                             dataIt->getProperty (".val").getTag ());
         }
      }
   }

   // --- validate groups
   auto groupIt  = foundGroups.getKeys ().begin ();
   auto matchIt = foundGroups.getValues ().begin ();
   for (; groupIt.isValid (); (++groupIt,++matchIt))  {
      auto chIt = schemaG->begin (*groupIt);
      const SxVariant &key = chIt->getProperty (".key");
      int64_t nItemsFound = 0;

      int64_t n = getInt (chIt, "nItems", max<int64_t>());
      int64_t maxItems = getInt (chIt, "maxItems", max<int64_t>());
      int64_t minItems = getInt (chIt, "minItems", min<int64_t>());

      for (auto mIt = matchIt->begin (); mIt != matchIt->end (); ++mIt)  {
         auto dchIt = dataG->begin (*mIt);
         if (   dchIt->getProperty (".key") == key
             && dchIt->getProperty (".type").getInt () == Type::Group)  {

            nItemsFound++;
            if (n != max<int64_t>() && nItemsFound > n)  {
                validationError (  "too many groups with key '"
                                + key.getString ()
                                + "', expected " + n
                                + " got " + nItemsFound,
                                chIt->getProperty (".val").getTag (),
                                dchIt->getProperty (".val").getTag ());
            }
            if (maxItems != max<int64_t>() && nItemsFound > maxItems)  {
               validationError (  "too many groups with key '"
                                + key.getString ()
                                + "', expected less or equal to "
                                + maxItems + " got " + nItemsFound,
                                chIt->getProperty (".val").getTag (),
                                dchIt->getProperty (".val").getTag ());
            }
            validateObj (chIt, &dchIt);
         }
      }

      bool isOptional = getBool (chIt, "optional", false);

      if (nItemsFound == 0 && isOptional)  continue;

      if (n != max<int64_t>() && nItemsFound < n)  {
         validationError (  "too few groups with key '"
                          + key.getString ()
                          + "' expected " + n
                          + " got " + nItemsFound,
                          chIt->getProperty (".val").getTag (),
                          dataIt->getProperty (".val").getTag ());
      }

      if (minItems != min<int64_t>() && nItemsFound < minItems)  {
         validationError (  "too few groups with key '"
                          + key.getString ()
                          + "' expected greater or equal to "
                          + minItems + " got " + nItemsFound,
                          chIt->getProperty (".val").getTag (),
                          dataIt->getProperty (".val").getTag ());
      }

   }

   return true;
}

bool SxSchema::validate (const SxPtr<SxGraph<SxGProps> > &dataG_)
{
   SX_CHECK (dataG_.getPtr ());

   dataG = dataG_;

   auto dIt = dataG->begin ();
   auto sIt = schemaG->begin ();

   // --- skip dummy root
   ++sIt; ++dIt;
   Type t = (Type)getInt (sIt, "type", Type::Undefined);
   if (t == Type::Group)  validateObj (sIt, &dIt);
   else                   validateArray (sIt, dIt);

   return true;
}

SxGraph<SxGProps>::Iterator
SxSchema::get (const SxGraph<SxGProps>::Iterator &it,
               const SxString &key,
               const SxVariant &val)
{
   SX_CHECK (key != "");
   SX_CHECK (it.isValid ());

   for (ssize_t i = 0; i < it.getSizeOut (); ++i)
   {
      auto chIt = it.out (i);
      if (chIt->hasProperty (key))  {
         if (!val.isInitialized () || chIt->getProperty (key) == val)  {
            return chIt;
         }
      }
   }
   return SxGraph<SxGProps>::Iterator ();
}

void SxSchema::simplifyInt (SxGraph<SxGProps>::Iterator &sIt)
{
   int64_t minVal = getInt (sIt, "min", min<int64_t>());
   int64_t maxVal = getInt (sIt, "max", max<int64_t>());

   sIt->getProperty (".val").setType (SxVariantType::Int);
   sIt->getProperty (".val").set (minVal);
   sIt->getProperty (".val").setLimits (minVal, maxVal);
   SX_CHECK (sIt->getProperty (".val").getType() == SxVariantType::Int);
}

void SxSchema::simplifyDouble (SxGraph<SxGProps>::Iterator &sIt)
{
   double minVal = getDouble (sIt, "min", min<double>());
   double maxVal = getDouble (sIt, "max", max<double>());

   sIt->getProperty (".val").setType (SxVariantType::Double);
   sIt->getProperty (".val").set (minVal);
   sIt->getProperty (".val").setLimits (minVal, maxVal);
   SX_CHECK (sIt->getProperty (".val").getType() == SxVariantType::Double);
}

void SxSchema::simplifyString (SxGraph<SxGProps>::Iterator &sIt)
{
   ssize_t minVal = (ssize_t)getInt (sIt, "min", min<ssize_t>());
   ssize_t maxVal = (ssize_t)getInt (sIt, "max", max<ssize_t>());

   sIt->getProperty (".val").setType (SxVariantType::String);
   sIt->getProperty (".val").set (SxString (""));
   sIt->getProperty (".val").setLimits (minVal, maxVal);
   SX_CHECK (sIt->getProperty (".val").getType() == SxVariantType::String);
}

void SxSchema::simplifyObj (SxGraph<SxGProps>::Iterator &sIt)
{
   for (ssize_t i = 0; i < sIt.getSizeOut (); ++i)  {
      auto chIt = sIt.out (i);

      if (chIt->getProperty (".type").getInt () == Type::Group)  {

         auto typeIt = get (chIt, ".key", "type");

         SxString typeStr = typeIt->getProperty (".val").toString ();

         Type elemType = (Type)SxParserAst::getTypeId (typeStr);
         typeIt->setProperty (".val", (int)elemType);

         if (elemType == Type::Group)  {
            simplifyObj (chIt);
         }  else if (elemType == Type::Int)  {
            simplifyInt (chIt);
         }  else if (elemType == Type::Double)  {
            simplifyDouble (chIt);
         }  else if (elemType == Type::String)  {
            simplifyString (chIt);
         }
      }

   }
}

void SxSchema::simplifySchemaAst ()
{
   SX_CHECK (schemaG.getPtr ());

   auto sIt = schemaG->begin ();
   // --- skip dummy root
   ++sIt;

   topLevelDefs = getBool (sIt, "topLevelDefs", false);

   allowUndeclared = getBool (sIt, "allowUndeclared", false);

   auto typeIt = get (sIt, ".key", "type");

   Type t = (Type)SxParserAst::getTypeId (typeIt->getProperty (".val").toString ());
   typeIt->setProperty (".val", (int)t);

   if (t == Type::Group)  simplifyObj (sIt);
}

